# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

# TODO: This the docstrings inside this module's set of functions
# can do without the `type`s.
# This is because the types are already annotated on the function.

import logging
from functools import partial
from typing import Any, Callable, Dict, List, Optional

import networkx as nx

from graphein.protein.features.sequence.utils import (
    aggregate_feature_over_chains,
    compute_feature_over_chains,
)
from graphein.protein.features.utils import (
    aggregate_graph_feature_over_chains,
    convert_graph_dict_feat_to_series,
)

log = logging.getLogger(__name__)


def compute_propy_feature(
    G: nx.Graph,
    func: Callable,
    feature_name: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Computes Propy Descriptors over chains in a Protein Graph

    :param G: Protein Graph
    :type G: nx.Graph
    :param func: ProPy wrapper function to compute
    :type func: Callable
    :param feature_name: Name of feature to index it in the nx.Graph object
    :type feature_name: str
    :param aggregation_type: Type of aggregation to use when aggregating a feature over multiple chains. One of:
        ["mean", "man", "sum"]. Defaults to None.
    :type aggregation_type: List[str], optional
    :return G: Returns protein Graph with features added. Features are accessible with G.graph[{feature_name}_{chain | aggegation_type}]
    :rtype: nx.Graph
    """
    G = compute_feature_over_chains(G, func, feature_name=feature_name)

    # Convert to Series
    G = convert_graph_dict_feat_to_series(G, feature_name=feature_name)

    # Aggregate features
    if len(G.graph["chain_ids"]) > 1:
        if aggregation_type:
            G = aggregate_graph_feature_over_chains(
                G, feature_name=feature_name, aggregation_type=aggregation_type
            )
    else:
        log.debug("Aggregation not carried out on single-chain graph")
    return G


def amino_acid_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of Amino acids for a given protein sequence.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use
    :type aggregation_type: Optional[List[str]]
    :return: Protein Graph with amino_acid_composition feature added. G.graph["amino_acid_composition_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateAAComposition

    func = CalculateAAComposition
    feature_name = "amino_acid_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def dipeptide_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of dipeptidefor a given protein sequence. Contains composition of 400 dipeptides

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with dipeptide_composition feature added. G.graph["dipeptide_composition_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateDipeptideComposition

    func = CalculateDipeptideComposition
    feature_name = "dipeptide_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def aa_dipeptide_composition(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the composition of AADs, dipeptide and 3-mers for a given protein sequence. Contains all composition
        values of AADs, dipeptide and 3-mers (8420).

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with aa_dipeptide_composition feature added.
        G.graph["aa_dipeptide_composition_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.AAComposition import CalculateAADipeptideComposition

    func = CalculateAADipeptideComposition
    feature_name = "aa_dipeptide_composition"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def aa_spectrum(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate the spectrum descriptors of 3-mers for a given protein. Contains the composition values of 8000 3-mers

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with aa_spectrum feature added. G.graph["aa_spectrum_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.AAComposition import GetSpectrumDict

    func = GetSpectrumDict
    feature_name = "aa_spectrum"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


######## CTD Funcs #######


def all_composition_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all composition descriptors based on seven different properties of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_descriptors feature added.
        G.graph["composition_descriptors_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "composition_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_ctd_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all CTD descriptors based seven different properties of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with ctd_descriptors feature added.
        G.graph["ctd_descriptors_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCTD

    func = CalculateCTD
    feature_name = "ctd_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute composition descriptors.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param AAProperty: contains classification of amino acids such as _Polarizability.
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_{AAPName} feature added.
        G.graph["composition_{AAPName}_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateComposition

    func = partial(
        CalculateComposition, AAProperty=AAProperty, AAPName=AAPName
    )
    feature_name = f"composition_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_charge feature added.
        G.graph["composition_charge_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionCharge

    func = CalculateCompositionCharge
    feature_name = "composition_charge"

    G = compute_feature_over_chains(G, func, feature_name="composition_charge")

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_hydrophobicity feature added.
        G.graph["composition_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionHydrophobicity

    func = CalculateCompositionHydrophobicity
    feature_name = "composition_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate composition descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_normalized_vdwv feature added.
        G.graph["composition_normalized_vdwv_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionNormalizedVDWV

    func = CalculateCompositionNormalizedVDWV
    feature_name = "composition_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_polarity feature added.
        G.graph["composition_polarity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionPolarity

    func = CalculateCompositionPolarity
    feature_name = "composition_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_polarizability feature added.
        G.graph["composition_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionPolarizability

    func = CalculateCompositionPolarizability
    feature_name = "composition_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_secondary_str feature added.
        G.graph["composition_secondary_str_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionSecondaryStr

    func = CalculateCompositionSecondaryStr
    feature_name = "composition_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def composition_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate composition descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with composition_solvent_accessibility feature added.
        G.graph["composition_solvent_accessibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateCompositionSolventAccessibility

    func = CalculateCompositionSolventAccessibility
    feature_name = "composition_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_distribution_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all distribution descriptors based on seven different properties of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_descriptors feature added.
        G.graph["distribution_descriptors_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "distribution_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute distribution descriptors.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param AAProperty: contains classifciation of amino acids such as _Polarizability.
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_{AAPName} feature added.
        G.graph["distribution_{AAPName}_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistribution

    func = partial(
        CalculateDistribution, AAProperty=AAProperty, AAPName=AAPName
    )
    feature_name = f"distribution_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_charge feature added.
        G.graph["distribution_charge_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionCharge

    func = CalculateDistributionCharge
    feature_name = "distribution_charge"

    G = compute_feature_over_chains(
        G, func, feature_name="distribution_charge"
    )

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_hydrophobicity feature added.
        G.graph["distribution_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionHydrophobicity

    func = CalculateDistributionHydrophobicity
    feature_name = "distribution_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_normalized_vdwv feature added.
        G.graph["distribution_normalized_vdwv_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionNormalizedVDWV

    func = CalculateDistributionNormalizedVDWV
    feature_name = "distribution_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_polarity feature added.
        G.graph["distribution_polarity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionPolarity

    func = CalculateDistributionPolarity
    feature_name = "distribution_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_polarizability feature added.
        G.graph["distribution_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionPolarizability

    func = CalculateDistributionPolarizability
    feature_name = "distribution_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_secondary_str feature added.
        G.graph["distribution_secondary_str_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionSecondaryStr

    func = CalculateDistributionSecondaryStr
    feature_name = "distribution_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def distribution_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate distribution descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with distribution_solvent_accessibility feature added.
        G.graph["distribution_solvent_accessibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateDistributionSolventAccessibility

    func = CalculateDistributionSolventAccessibility
    feature_name = "distribution_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def all_transition_descriptors(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate all transition descriptors based on seven different properties of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_descriptors feature added.
        G.graph["transition_descriptors_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateC

    func = CalculateC
    feature_name = "transition_descriptors"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_descriptor(
    G: nx.Graph,
    AAProperty: Dict[Any, Any],
    AAPName: str,
    aggregation_type: Optional[List[str]] = None,
) -> nx.Graph:
    """
    Compute transition descriptors.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param AAProperty: contains classifciation of amino acids such as _Polarizability.
    :type AAProperty: Dict[Any, Any]
    :param AAPName: used for indicating a AAP name.
    :type AAPName: str
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_{AAPName} feature added.
        G.graph["transition_{AAPName}_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransition

    func = partial(CalculateTransition, AAProperty=AAProperty, AAPName=AAPName)
    feature_name = f"transition_{AAPName}"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_charge(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate transition descriptors based on Charge of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_charge feature added.
        G.graph["transition_charge_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionCharge

    func = CalculateTransitionCharge
    feature_name = "transition_charge"

    G = compute_feature_over_chains(G, func, feature_name="transition_charge")

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Calculate transition descriptors based on Hydrophobicity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_hydrophobicity feature added.
        G.graph["transition_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionHydrophobicity

    func = CalculateTransitionHydrophobicity
    feature_name = "transition_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_normalized_vdwv(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on NormalizedVDWV of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_normalized_vdwv feature added.
        G.graph["transition_normalized_vdwv_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionNormalizedVDWV

    func = CalculateTransitionNormalizedVDWV
    feature_name = "transition_normalised_vdwv"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_polarity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on Polarity of AADs.

    :param G: Protein Graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_polarity feature added.
        G.graph["transition_polarity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionPolarity

    func = CalculateTransitionPolarity
    feature_name = "transition_polarity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on Polarizability of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_polarizability feature added.
        G.graph["transition_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionPolarizability

    func = CalculateTransitionPolarizability
    feature_name = "transition_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_secondary_str(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on SecondaryStr of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_secondary_str feature added.
        G.graph["transition_secondary_str_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionSecondaryStr

    func = CalculateTransitionSecondaryStr
    feature_name = "transition_secondary_str"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def transition_solvent_accessibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate transition descriptors based on SolventAccessibility of AADs.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with transition_solvent_accessibility feature added.
        G.graph["transition_solvent_accessibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.CTD import CalculateTransitionSolventAccessibility

    func = CalculateTransitionSolventAccessibility
    feature_name = "transition_solvent_accessibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


####### Autocorrelation descriptors ########


def autocorrelation_total(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute all autocorrelation descriptors based on 8 properties of AADs. result contains 30*8*3=720 normalized Moreau
        Broto, Moran, and Geary

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_total feature added.
        G.graph["autocorrelation_total_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateAutoTotal

    func = CalculateAutoTotal
    feature_name = "autocorrelation_total"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute Geary autocorrelation descriptors based on 8 properties of AADs. Result contains 30*8=240 Geary
        autocorrelation descriptors based on the given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_all feature added.
        G.graph["autocorrelation_geary_all_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoTotal

    func = CalculateGearyAutoTotal
    feature_name = "autocorrelation_geary_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on AvFlexibility. contains 30 Geary Autocorrelation
        descriptors based on AvFlexibility.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_av_flexibility feature added.
        G.graph["autocorrelation_geary_av_flexibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoAvFlexibility

    func = CalculateGearyAutoAvFlexibility
    feature_name = "autocorrelation_geary_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on FreeEnergy. result contains 30 Geary Autocorrelation
        descriptors based on FreeEnergy.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_free_energy feature added.
        G.graph["autocorrelation_geary_av_free_energy_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoFreeEnergy

    func = CalculateGearyAutoFreeEnergy
    feature_name = "autocorrelation_geary_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on hydrophobicity. result contains 30 Geary Autocorrelation
        descriptors based on hydrophobicity.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_hydrophobicity feature added.
        G.graph["autocorrelation_geary_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoHydrophobicity

    func = CalculateGearyAutoHydrophobicity
    feature_name = "autocorrelation_geary_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on Mutability. result contains 30 Geary Autocorrelation
        descriptors based on mutability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_mutability feature added.
        G.graph["autocorrelation_geary_mutability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoMutability

    func = CalculateGearyAutoMutability
    feature_name = "autocorrelation_geary_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on polarizability. result contains 30 Geary Autocorrelation
        descriptors based on polarizability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_polarizability feature added.
        G.graph["autocorrelation_geary_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoPolarizability

    func = CalculateGearyAutoPolarizability
    feature_name = "autocorrelation_geary_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on ResidueASA. result contains 30 Geary Autocorrelation
        descriptors based on ResidueASA.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_residue_asa feature added.
        G.graph["autocorrelation_geary_residue_asa_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoResidueASA

    func = CalculateGearyAutoResidueASA
    feature_name = "autocorrelation_geary_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on ResidueVol. result contains 30 Geary Autocorrelation
        descriptors based on ResidueVol.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_residue_vol feature added.
        G.graph["autocorrelation_geary_residue_vol_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoResidueVol

    func = CalculateGearyAutoResidueVol
    feature_name = "autocorrelation_geary_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_geary_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Geary Autocorrelation descriptors based on Steric. result contains 30 Geary Autocorrelation
        descriptors based on Steric

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_geary_steric feature added.
        G.graph["autocorrelation_geary_steric_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateGearyAutoSteric

    func = CalculateGearyAutoSteric()
    feature_name = "autocorrelation_geary_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute Moran autocorrelation descriptors based on 8 properties of AADs. Result contains 30*8=240 Moran
        autocorrelation descriptors based on the given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_all feature added.
        G.graph["autocorrelation_moran_all_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoTotal

    func = CalculateMoranAutoTotal
    feature_name = "autocorrelation_moran_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on AvFlexibility. contains 30 Moran Autocorrelation
        descriptors based on AvFlexibility.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_av_flexibility feature added.
        G.graph["autocorrelation_moran_av_flexibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoAvFlexibility

    func = CalculateMoranAutoAvFlexibility
    feature_name = "autocorrelation_moran_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on FreeEnergy. result contains 30 Moran Autocorrelation
        descriptors based on FreeEnergy.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_free_energy feature added.
        G.graph["autocorrelation_moran_av_free_energy_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoFreeEnergy

    func = CalculateMoranAutoFreeEnergy
    feature_name = "autocorrelation_moran_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on hydrophobicity. result contains 30 Moran Autocorrelation
        descriptors based on hydrophobicity.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_hydrophobicity feature added.
        G.graph["autocorrelation_moran_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoHydrophobicity

    func = CalculateMoranAutoHydrophobicity
    feature_name = "autocorrelation_moran_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on Mutability. result contains 30 Moran Autocorrelation
        descriptors based on mutability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_mutability feature added.
        G.graph["autocorrelation_moran_mutability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoMutability

    func = CalculateMoranAutoMutability
    feature_name = "autocorrelation_moran_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on polarizability. result contains 30 Moran Autocorrelation
        descriptors based on polarizability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_polarizability feature added.
        G.graph["autocorrelation_moran_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoPolarizability

    func = CalculateMoranAutoPolarizability
    feature_name = "autocorrelation_moran_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on ResidueASA. result contains 30 Moran Autocorrelation
        descriptors based on ResidueASA.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_residue_asa feature added.
        G.graph["autocorrelation_moran_residue_asa_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoResidueASA

    func = CalculateMoranAutoResidueASA
    feature_name = "autocorrelation_moran_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on ResidueVol. result contains 30 Moran Autocorrelation
        descriptors based on ResidueVol.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_residue_vol feature added.
        G.graph["autocorrelation_moran_residue_vol_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoResidueVol

    func = CalculateMoranAutoResidueVol
    feature_name = "autocorrelation_moran_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_moran_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the Moran Autocorrelation descriptors based on Steric. result contains 30 Moran Autocorrelation
        descriptors based on Steric

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_moran_steric feature added.
        G.graph["autocorrelation_moran_steric_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateMoranAutoSteric

    func = CalculateMoranAutoSteric()
    feature_name = "autocorrelation_moran_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_all(
    G: nx.Graph, aggregation_type: Optional[List[str]] = None
) -> nx.Graph:
    """
    Compute NormalizedMoreauBroto autocorrelation descriptors based on 8 properties of AADs. Result contains 30*8=240
        NormalizedMoreauBroto autocorrelation descriptors based on the given properties(i.e., _AAPropert).

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_all feature added.
        G.graph["autocorrelation_normalized_moreau_broto_all_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateNormalizedMoreauBrotoAutoTotal

    func = CalculateNormalizedMoreauBrotoAutoTotal
    feature_name = "autocorrelation_normalized_moreau_broto_all"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_av_flexibility(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on AvFlexibility. contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on AvFlexibility.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_av_flexibility feature added.
        G.graph["autocorrelation_normalized_moreau_broto_av_flexibility_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoAvFlexibility,
    )

    func = CalculateNormalizedMoreauBrotoAutoAvFlexibility
    feature_name = "autocorrelation_normalized_moreau_broto_av_flexibility"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_free_energy(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on FreeEnergy. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on FreeEnergy.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_free_energy feature added.
        G.graph["autocorrelation_normalized_moreau_broto_av_free_energy_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoFreeEnergy,
    )

    func = CalculateNormalizedMoreauBrotoAutoFreeEnergy
    feature_name = "autocorrelation_normalized_moreau_broto_free_energy"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_hydrophobicity(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on hydrophobicity. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on hydrophobicity.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_hydrophobicity feature added.
        G.graph["autocorrelation_normalized_moreau_broto_hydrophobicity_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoHydrophobicity,
    )

    func = CalculateNormalizedMoreauBrotoAutoHydrophobicity
    feature_name = "autocorrelation_normalized_moreau_broto_hydrophobicity"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_mutability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on Mutability. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on mutability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_mutability feature added.
        G.graph["autocorrelation_normalized_moreau_broto_mutability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoMutability,
    )

    func = CalculateNormalizedMoreauBrotoAutoMutability
    feature_name = "autocorrelation_normalized_moreau_broto_mutability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_polarizability(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on polarizability. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on polarizability.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_polarizability feature added.
        G.graph["autocorrelation_normalized_moreau_broto_polarizability_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoPolarizability,
    )

    func = CalculateNormalizedMoreauBrotoAutoPolarizability
    feature_name = "autocorrelation_normalized_moreau_broto_polarizability"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_residue_asa(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on ResidueASA. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on ResidueASA.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_residue_asa feature added.
        G.graph["autocorrelation_normalized_moreau_broto_residue_asa_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoResidueASA,
    )

    func = CalculateNormalizedMoreauBrotoAutoResidueASA
    feature_name = "autocorrelation_normalized_moreau_broto_residue_asa"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_residue_vol(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on ResidueVol. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on ResidueVol.

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_residue_vol feature added.
        G.graph["autocorrelation_normalized_moreau_broto_residue_vol_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import (
        CalculateNormalizedMoreauBrotoAutoResidueVol,
    )

    func = CalculateNormalizedMoreauBrotoAutoResidueVol
    feature_name = "autocorrelation_normalized_moreau_broto_residue_vol"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


def autocorrelation_normalized_moreau_broto_steric(
    G: nx.Graph, aggregation_type: Optional[List[str]]
) -> nx.Graph:
    """
    Calculate the NormalizedMoreauBroto Autocorrelation descriptors based on Steric. result contains 30
        NormalizedMoreauBroto Autocorrelation descriptors based on Steric

    :param G: Protein graph to featurise
    :type G: nx.Graph
    :param aggregation_type: Aggregation types to use over chains
    :type aggregation_type: List[Optional[str]]
    :return: Protein Graph with autocorrelation_normalized_moreau_broto_steric feature added.
        G.graph["autocorrelation_normalized_moreau_broto_steric_{chain | aggregation_type}"]
    :rtype: nx.Graph
    """
    from propy.Autocorrelation import CalculateNormalizedMoreauBrotoAutoSteric

    func = CalculateNormalizedMoreauBrotoAutoSteric()
    feature_name = "autocorrelation_normalized_moreau_broto_steric"

    return compute_propy_feature(
        G,
        func=func,
        feature_name=feature_name,
        aggregation_type=aggregation_type,
    )


####### Quasi Sequence Order Descriptors ######


def quasi_sequence_order_aa_composition(G: nx.Graph) -> nx.Graph:
    from propy.QuasiSequenceOrder import GetAAComposition

    func = GetAAComposition
    G = compute_feature_over_chains(
        G, func, feature_name="quasi_seq_order_aa_composition"
    )

    return G


def quasi_sequence_order(
    G: nx.Graph, maxlag: int = 30, weight: float = 0.1
) -> nx.Graph:
    """
    Compute quasi-sequence-order descriptors for a given protein.

    Kuo-Chen Chou. Prediction of Protein Subcellar Locations by Incorporating Quasi-Sequence-Order Effect.
        Biochemical and Biophysical Research Communications 2000, 278, 477-483.

    :param maxlag: (int, optional (default: 30)) – the maximum lag and the length of the protein should be larger than
        maxlag
    :type maxlag: int
    :param weight: (float, optional (default: 0.1)) – a weight factor. Please see reference 1 for its choice.
    :type weight: float
    :returns:
    :rtype: nx.Graph
    """
    from propy.QuasiSequenceOrder import GetQuasiSequenceOrder

    func = partial(GetQuasiSequenceOrder, weight=weight, maxlag=maxlag)
    G = compute_feature_over_chains(
        G, func, feature_name="quasi_sequence_order"
    )
    return G


def sequence_order_coupling_number_total(
    G: nx.Graph, maxlag: int = 30
) -> nx.Graph:
    """
    Compute the sequence order coupling numbers from 1 to maxlag for a given protein sequence.

    :param G: Protein Graph
    :type G: nx.Graph
    :param maxlag (int, optional (default: 30)) – the maximum lag and the length of the protein should be larger
    :type maxlag: int
    :returns:
    :rtype: nx.Graph
    """
    from propy.QuasiSequenceOrder import GetSequenceOrderCouplingNumberTotal

    func = partial(GetSequenceOrderCouplingNumberTotal, maxlag=maxlag)
    G = compute_feature_over_chains(
        G, func, feature_name="sequence_order_coupling_number_total"
    )

    return G


# Todo feature aggregation

if __name__ == "__main__":
    pass



"""Featurization functions for amino acids."""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein

import logging
from functools import lru_cache
from pathlib import Path
from typing import List, Optional, Union

import numpy as np
import pandas as pd

log = logging.getLogger(__name__)


@lru_cache()
def load_expasy_scales() -> pd.DataFrame:
    """
    Load pre-downloaded EXPASY scales.

    This helps with node featuarization.

    The function is LRU-cached in memory for fast access
    on each function call.

    :returns: pd.DataFrame containing expasy scales
    :rtype: pd.DataFrame
    """
    fpath = Path(__file__).parent / "amino_acid_properties.csv"
    log.debug(f"Reading Expasy protein scales from: {fpath}")
    df = pd.read_csv(fpath, index_col=0)
    return df


@lru_cache()
def load_meiler_embeddings() -> pd.DataFrame:
    """
    Load pre-downloaded Meiler embeddings.

    This helps with node featurization.

    The function is LRU-cached in memory for fast access
    on each function call.

    :returns: pd.DataFrame containing Meiler Embeddings from Meiler et al. 2001
    :rtype: pd.DataFrame
    """
    fpath = Path(__file__).parent / "meiler_embeddings.csv"
    log.debug(f"Reading meiler embeddings from: {fpath}")
    df = pd.read_csv(fpath, index_col=0)
    return df


def expasy_protein_scale(
    n, d, selection: Optional[List[str]] = None, return_array: bool = False
) -> Union[pd.Series, np.ndarray]:
    """
    Return amino acid features that come from the EXPASY protein scale.

    Source: https://web.expasy.org/protscale/

    :param n: Node in a NetworkX graph
    :param d: NetworkX node attributes.
    :param selection: List of columns to select. Viewable in graphein.protein.features.nodes.meiler_embeddings
    :type selection: List[str], optional
    :param return_array: Bool indicating whether or not to return a np.ndarray of the features. Default is pd.Series
    :type return_array: bool
    :returns: pd.Series of amino acid features
    :rtype: pd.Series
    """
    df = load_expasy_scales()
    amino_acid = d["residue_name"]
    try:
        features = df[amino_acid]
        if selection is not None:
            features = features.filter(selection)
    except:
        features = pd.Series(np.zeros(len(df)))

    if return_array:
        features = np.array(features)

    d["expasy"] = features

    return features


def meiler_embedding(
    n, d, return_array: bool = False
) -> Union[pd.Series, np.array]:
    """
    Return amino acid features from reduced dimensional embeddings of amino acid physicochemical properties.

    Source: https://link.springer.com/article/10.1007/s008940100038
    doi: https://doi.org/10.1007/s008940100038

    :param n: Node in a NetworkX graph
    :param d: NetworkX node attributes.
    :returns: pd.Series of amino acid features
    :rtype: pd.Series
    """
    df = load_meiler_embeddings()
    amino_acid = d["residue_name"]
    try:
        features = df[amino_acid]
    except:
        features = pd.Series(np.zeros(len(df)))

    if return_array:
        features = np.array(features)

    d["meiler"] = features

    return features



from bioservices import HGNC, UniProt


def add_sequence_to_nodes(n, d):
    """
    Maps UniProt ACC to UniProt ID. Retrieves sequence from UniProt and adds it to the node

    :param n: Graph node.
    :param d: Graph attribute dictionary.
    """

    h = HGNC(verbose=False)
    u = UniProt(verbose=False)

    d["uniprot_ids"] = h.fetch("symbol", d["gene_id"])["response"]["docs"][0][
        "uniprot_ids"
    ]

    # Todo these API calls should probably be batched
    # Todo mapping with bioservices to support other protein IDs?

    for id in d["uniprot_ids"]:
        d[f"sequence_{id}"] = u.get_fasta_sequence(id)



import functools

# %%
# Graphein
# Author: Ramon Vinas, Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
import logging
import os
import zipfile
from pathlib import Path
from typing import Callable, List, Optional

import pandas as pd
import wget

from graphein.utils.utils import filter_dataframe

log = logging.getLogger(__name__)


def _download_RegNetwork(
    root_dir: Optional[Path] = None, network_type: str = "human"
) -> str:
    """
    Downloads RegNetwork regulatory interactions to the root directory. Returns the filepath.

    :param root_dir: Path object specifying the location to download RegNetwork to. Default is None which downloads to the dataset/ directory inside graphein.
    :type root_dir: patlib.Path, optional
    :param network_type: Specifies whether to download human or mouse regulatory network. Supported values: "human" (default), "mouse".
    :type network_type: str
    :returns: path to downloaded RegNetwork
    :rtype: str
    """
    human_url = "http://www.regnetworkweb.org/download/human.zip"
    mouse_url = "http://regnetworkweb.org/download/mouse.zip"

    if network_type == "human":
        url = human_url
    elif network_type == "mouse":
        url = mouse_url
    else:
        raise ValueError(
            f"network_type: {network_type} is unsupported. Please use 'human' or 'mouse'"
        )

    # If no root dir is provided, use the dataset directory inside graphein.
    if root_dir is None:
        root_dir = Path(__file__).parent.parent.parent / "datasets"

    regnetwork_dir = f"{root_dir}/regnetwork"
    Path(regnetwork_dir).mkdir(parents=False, exist_ok=True)
    compressed_file = f"{regnetwork_dir}/human.zip"
    out_dir = f"{regnetwork_dir}/human"
    file = f"{out_dir}/human.source"

    # Download data and unzip
    if not os.path.exists(file):
        log.info("Downloading RegNetwork ...")
        wget.download(url, compressed_file)

        with zipfile.ZipFile(compressed_file, "r") as zip_ref:
            zip_ref.extractall(out_dir)

    return file


def _download_RegNetwork_regtypes(root_dir: Optional[Path] = None) -> str:
    """
    Downloads RegNetwork regulatory interactions types to the root directory

    :param root_dir: Path object specifying the location to download RegNetwork to
    """
    url = "http://www.regnetworkweb.org/download/RegulatoryDirections.zip"

    if root_dir is None:
        root_dir = Path(__file__).parent.parent.parent / "datasets"

    regnetwork_dir = f"{root_dir}/regnetwork"
    Path(regnetwork_dir).mkdir(parents=False, exist_ok=True)
    compressed_file = f"{regnetwork_dir}/RegulatoryDirections.zip"
    out_dir = f"{regnetwork_dir}/human"
    file = f"{out_dir}/new_kegg.human.reg.direction.txt"

    # Download data and unzip
    if not os.path.exists(file):
        log.info("Downloading RegNetwork reg types ...")
        wget.download(url, compressed_file)

        with zipfile.ZipFile(compressed_file, "r") as zip_ref:
            zip_ref.extractall(out_dir)

    return file


@functools.lru_cache()
def load_RegNetwork_interactions(
    root_dir: Optional[Path] = None,
) -> pd.DataFrame:
    """
    Loads RegNetwork interaction datafile. Downloads the file first if not already present.
    """
    file = _download_RegNetwork(root_dir)
    return pd.read_csv(
        file, delimiter="\t", header=None, names=["g1", "id1", "g2", "id2"]
    )


@functools.lru_cache()
def load_RegNetwork_regulation_types(
    root_dir: Optional[Path] = None,
) -> pd.DataFrame:
    """
    Loads RegNetwork regulation types. Downloads the file first if not already present.
    """
    file = _download_RegNetwork_regtypes(root_dir)
    return pd.read_csv(
        file,
        delimiter=" ",
        header=None,
        names=["tf", "id1", "target", "id2", "regtype"],
        skiprows=1,
    )


def parse_RegNetwork(
    gene_list: List[str], root_dir: Optional[Path] = None
) -> pd.DataFrame:
    """
    Parser for RegNetwork interactions

    :param gene_list: List of gene identifiers
    :return Pandas dataframe with the regulatory interactions between genes in the gene list
    """
    # Load dataframes
    df = load_RegNetwork_interactions(root_dir)
    reg_df = load_RegNetwork_regulation_types(root_dir)

    df = pd.merge(
        df,
        reg_df,
        how="outer",
        left_on=["g1", "g2"],
        right_on=["tf", "target"],
    )
    df["g1"] = df["g1"].combine_first(df["tf"])
    df["g2"] = df["g2"].combine_first(df["target"])

    # Select input genes
    df = df[df["g1"].isin(gene_list) & df["g2"].isin(gene_list)]

    return df


def filter_RegNetwork(
    df: pd.DataFrame, funcs: Optional[List[Callable]] = None
) -> pd.DataFrame:
    """
    Filters results of RegNetwork call by providing a list of user-defined functions that accept a dataframe and return a dataframe

    :param df: pd.Dataframe to filter
    :param funcs: list of functions that carry out dataframe processing
    :return: processed dataframe
    """
    if funcs is not None:
        df = filter_dataframe(df, funcs)

    return df


def standardise_RegNetwork(df: pd.DataFrame) -> pd.DataFrame:
    """
    Standardises STRING dataframe, e.g. puts everything into a common format

    :param df: Source specific Pandas dataframe
    :type df: pd.DataFrame
    :return: Standardised dataframe
    :rtype: pd.DataFrame
    """
    # Rename & delete columns
    df = df[["g1", "g2", "regtype"]]

    # Add source column
    df["source"] = "RegNetwork"

    # Standardise regulatory types
    df["regtype"].replace(
        {"-->": "+", "--|": "-", None: "?", "-p": "?", "-/-": "?"},
        inplace=True,
    )

    return df


def RegNetwork_df(
    gene_list: List[str],
    root_dir: Optional[Path] = None,
    filtering_funcs: Optional[List[Callable]] = None,
) -> pd.DataFrame:
    """
    Generates standardised dataframe with RegNetwork protein-protein interactions, filtered according to user's input
    :return: Standardised dataframe with RegNetwork interactions
    """
    df = parse_RegNetwork(gene_list=gene_list, root_dir=root_dir)
    df = filter_RegNetwork(df, filtering_funcs)
    df = standardise_RegNetwork(df)

    return df


if __name__ == "__main__":
    df = RegNetwork_df(["AATF", "MYC", "USF1", "SP1", "TP53", "DUSP1"])
    print(df)



"""Functions for working with RNA Secondary Structure Graphs"""
# %%
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Emmanuele Rossi, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
import logging
from typing import Callable, Dict, List, Optional

import networkx as nx

from graphein.utils.utils import (
    annotate_edge_metadata,
    annotate_graph_metadata,
    annotate_node_metadata,
    compute_edges,
)

log = logging.getLogger(__name__)

RNA_BASES: List[str] = ["A", "U", "G", "C", "I"]

RNA_BASE_COLORS: Dict[str, str] = {
    "A": "r",
    "U": "b",
    "G": "g",
    "C": "y",
    "I": "m",
}

CANONICAL_BASE_PAIRINGS: Dict[str, str] = {
    "A": ["U"],
    "U": ["A"],
    "G": ["C"],
    "C": ["G"],
}

WOBBLE_BASE_PAIRINGS: Dict[str, str] = {
    "A": ["I"],
    "U": ["G", "I"],
    "G": ["U"],
    "C": ["I"],
    "I": ["A", "C", "U"],
}

VALID_BASE_PAIRINGS = {
    key: CANONICAL_BASE_PAIRINGS.get(key, [])
    + WOBBLE_BASE_PAIRINGS.get(key, [])
    for key in set(
        list(CANONICAL_BASE_PAIRINGS.keys())
        + list(WOBBLE_BASE_PAIRINGS.keys())
    )
}

SIMPLE_DOTBRACKET_NOTATION = ["(", ".", ")"]
SUPPORTED_PSEUDOKNOT_NOTATION = ["[", "]", "{", "}", "<", ">"]
SUPPORTED_DOTBRACKET_NOTATION = (
    SIMPLE_DOTBRACKET_NOTATION + SUPPORTED_PSEUDOKNOT_NOTATION
)


def validate_rna_sequence(s: str) -> None:
    """
    Validate RNA sequence. This ensures that it only containts supported bases. Supported bases are: "A", "U", "G", "C", "I".
    Supported bases can be accessed in graphein.rna.graphs.RNA_BASES

    :param s: Sequence to validate
    :type s: str
    :raises ValueError: Raises ValueError if the sequence contains an unsupported base character
    """
    letters_used = set(s)
    if not letters_used.issubset(RNA_BASES):
        offending_letter = letters_used.difference(RNA_BASES)
        position = s.index(offending_letter)
        raise ValueError(
            f"Invalid letter {offending_letter} found at position {position} in the sequence {s}."
        )


def validate_lengths(db: str, seq: str) -> None:
    """
    Check lengths of dotbracket and sequence match

    :param db: Dotbracket string to check
    :type db: str
    :param seq: RNA nucleotide sequence to check.
    :type seq: str
    :raises ValueError: Raises ValueError if lengths of dotbracket and sequence do not match.
    """
    if len(db) != len(seq):
        raise ValueError(
            f"Length of dotbracket ({len(db)}) does not match length of sequence ({len(seq)})."
        )


def validate_dotbracket(db: str) -> str:
    """
    Sanitize dotbracket string. This ensures that it only has supported symbols.
    SIMPLE_DOTBRACKET_NOTATION = ["(", ".", ")"]
    SUPPORTED_PSEUDOKNOT_NOTATION = ["[", "]", "{", "}", "<", ">"]
    SUPPORTED_DOTBRACKET_NOTATION = (
        SIMPLE_DOTBRACKET_NOTATION + SUPPORTED_PSEUDOKNOT_NOTATION
    )

    :param db: Dotbrack notation string
    :type db: str
    :raises ValueError: Raises ValueError if dotbracket notation contains unsupported symbols
    """
    chars_used = set(db)
    if not chars_used.issubset(SUPPORTED_DOTBRACKET_NOTATION):
        offending_letter = chars_used.difference(SUPPORTED_DOTBRACKET_NOTATION)
        position = db.index(offending_letter)
        raise ValueError(
            f"Invalid letter {offending_letter} found at position {position} in the sequence {db}."
        )


def construct_rna_graph(
    dotbracket: Optional[str],
    sequence: Optional[str],
    edge_construction_funcs: List[Callable],
    edge_annotation_funcs: Optional[List[Callable]] = None,
    node_annotation_funcs: Optional[List[Callable]] = None,
    graph_annotation_funcs: Optional[List[Callable]] = None,
) -> nx.Graph:
    """
    Constructs an RNA secondary structure graph from dotbracket notation

    :param dotbracket: Dotbracket notation representation of secondary structure
    :type dotbracket: str, optional
    :param sequence: Corresponding sequence RNA bases
    :type sequence: str, optional
    :param edge_construction_funcs: List of edge construction functions. Defaults to None.
    :type edge_construction_funcs: List[Callable], optional
    :param edge_annotation_funcs: List of edge metadata annotation functions. Defaults to None.
    :type edge_annotation_funcs: List[Callable], optional
    :param node_annotation_funcs: List of node metadata annotation functions. Defaults to None.
    :type node_annotation_funcs: List[Callable], optional
    :param graph_annotation_funcs: List of graph metadata annotation functions. Defaults to None
    :type graph_annotation_funcs: List[Callable], optional
    :return: nx.Graph of RNA secondary structure
    :rtype: nx.Graph
    """
    G = nx.Graph()

    # Build node IDs first.
    node_ids = (
        list(range(len(sequence)))
        if sequence
        else list(range(len(dotbracket)))
    )

    # Check sequence and dotbracket lengths match
    if dotbracket and sequence:
        validate_lengths(dotbracket, sequence)

    # add nodes
    G.add_nodes_from(node_ids)
    log.debug(f"Added {len(node_ids)} nodes")

    # Add dotbracket symbol if dotbracket is provided
    if dotbracket:
        validate_dotbracket(dotbracket)
        G.graph["dotbracket"] = dotbracket

        nx.set_node_attributes(
            G,
            dict(zip(node_ids, dotbracket)),
            "dotbracket_symbol",
        )

    # Add nucleotide base info if sequence is provided
    if sequence:
        validate_rna_sequence(sequence)
        G.graph["sequence"] = sequence
        nx.set_node_attributes(G, dict(zip(node_ids, sequence)), "nucleotide")
        colors = [RNA_BASE_COLORS[i] for i in sequence]
        nx.set_node_attributes(G, dict(zip(node_ids, colors)), "color")

    # Annotate additional graph metadata
    if graph_annotation_funcs is not None:
        G = annotate_graph_metadata(G, graph_annotation_funcs)

    # Annotate additional node metadata
    if node_annotation_funcs is not None:
        G = annotate_node_metadata(G, node_annotation_funcs)

    # Add edges
    G = compute_edges(G, edge_construction_funcs)

    # Annotate additional edge metadata
    if edge_annotation_funcs is not None:
        G = annotate_edge_metadata(G, edge_annotation_funcs)

    return G


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    from graphein.rna.edges import (
        add_all_dotbracket_edges,
        add_base_pairing_interactions,
        add_phosphodiester_bonds,
        add_pseudoknots,
    )

    edge_funcs_1 = [
        add_base_pairing_interactions,
        add_phosphodiester_bonds,
        add_pseudoknots,
    ]
    edge_funcs_2 = [add_all_dotbracket_edges]

    # g = construct_rna_graph(
    #    "((((....))))..(())",
    #    "AUGAUGAUGAUGCICIAU",
    #    edge_construction_funcs=edge_funcs_1,
    # )

    g = construct_rna_graph(
        "......((((((......[[[))))))......]]]....",
        sequence=None,
        edge_construction_funcs=edge_funcs_1,
    )

    """
    h = construct_rna_graph(
        "((((....))))..(())",
        "AUGAUGAUGAUGCICIAU",
        edge_construction_funcs=edge_funcs_2,
    )
    """

    # assert g.edges() == h.edges()

    nx.info(g)

    edge_colors = nx.get_edge_attributes(g, "color").values()
    node_colors = nx.get_node_attributes(g, "color").values()

    nx.draw(
        g, edge_color=edge_colors  # , node_color=node_colors, with_labels=True
    )
    plt.show()



"""Functions to create protein meshes via pymol."""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import importlib.util
import logging
from typing import List, NamedTuple, Optional, Tuple

from graphein.protein.config import ProteinMeshConfig
from graphein.utils.pymol import MolViewer
from graphein.utils.utils import import_message

try:
    from pytorch3d.structures import Meshes
except ImportError:
    import_message(
        submodule="graphein.protein.meshes",
        package="pytorch3d",
        conda_channel="pytorch3d",
    )


log = logging.getLogger(__name__)


def check_for_pymol_installation():
    """Checks for presence of a pymol installation"""
    spec = importlib.util.find_spec("pymol")
    if spec is None:
        log.error(
            "Please install pymol: conda install -c schrodinger pymol or conda install -c tpeulen pymol-open-source"
        )


def configure_pymol_session(
    config: Optional[ProteinMeshConfig] = None,
):
    """
    Configures a PyMol session based on config.parse_pymol_commands. Uses default parameters -"cKq".
    See: https://pymolwiki.org/index.php/Command_Line_Options

    :param config: ProteinMeshConfig to use
    :type config: graphein.protein.config.ProteinMeshConfig
    """
    pymol = MolViewer()

    # If no config is provided, use default
    if config is None:
        config = ProteinMeshConfig()

    # Start PyMol session
    pymol.start([config.pymol_command_line_options])


def get_obj_file(
    pdb_file: Optional[str] = None,
    pdb_code: Optional[str] = None,
    out_dir: Optional[str] = None,
    config: Optional[ProteinMeshConfig] = None,
) -> str:
    """
    Runs PyMol to compute surface/mesh for a given protein

    :param pdb_file:  path to pdb_file to use
    :type pdb_file: str, optional
    :param pdb_code: 4-letter pdb accession code
    :type pdb_code: str, optional
    :param out_dir: path to output. Defaults to /tmp/
    :type out_dir: str, optional
    :param config: ProteinMeshConfig containing pymol commands to run. Default is "show surface"
    :type config: graphein.protein.config.ProteinMeshConfig
    :return: returns path to .obj file (str)
    :rtype: str
    """
    pymol = MolViewer()

    check_for_pymol_installation()

    # Check inputs
    if not pdb_code and not pdb_file:
        raise ValueError("Please pass either a pdb_file or pdb_code argument")
    if pdb_code and pdb_file:
        raise ValueError(
            "Please pass either a pdb_file or pdb_code argument. Not both"
        )

    if out_dir is None:
        out_dir = "/tmp/"

    configure_pymol_session()

    # Load structure
    pymol.load(pdb_file) if pdb_file else pymol.fetch(pdb_code)
    # Create file_name
    file_name = (
        pdb_file[:-3] + "obj" if pdb_file else out_dir + pdb_code + ".obj"
    )

    if config is None:
        config = ProteinMeshConfig()

    commands = parse_pymol_commands(config)
    print(commands)
    run_pymol_commands(commands)

    # Save surface object as .obj
    pymol.do(f"save {file_name}")
    log.debug(f"Saved {file_name}")
    return file_name


def parse_pymol_commands(config: ProteinMeshConfig) -> List[str]:
    """
    Parses pymol commands from config. At the moment users can only supply a list of string commands.

    :param config: ProteinMeshConfig c
    :type config: ProteinMeshConfig
    :return: list of pymol commands to run
    :rtype: List[str]
    """
    if config is None:
        config = ProteinMeshConfig()

    # Todo parsing of individual pymol mesh calculation parameters. There's a lot of them so this is not a priority now.
    if config.pymol_commands is not None:
        return config.pymol_commands


def run_pymol_commands(commands: List[str]) -> None:
    """
    Runs Pymol Commands

    :param commands: List of commands to pass to PyMol
    :type commands: List[str]
    """
    pymol = MolViewer()

    for c in commands:
        log.debug(c)
        pymol.do(c)


def create_mesh(
    pdb_file: Optional[str] = None,
    pdb_code: Optional[str] = None,
    out_dir: Optional[str] = None,
    config: Optional[ProteinMeshConfig] = None,
) -> Tuple[torch.FloatTensor, NamedTuple, NamedTuple]:
    """
    Creates a PyTorch 3D mesh from a pdb_file or pdb code.

    :param pdb_file: path to pdb pdb_file. Defaults to None
    :type pdb_file: str, optional
    :param pdb_code: 4-letter PDB accession code. Defaults to None.
    :type pdb_code: str, optional
    :param out_dir: output directory to store ".obj" file. Defaults to /tmp/
    :type out_dir: str, optional
    :param config:  ProteinMeshConfig config to use. Defaults to default config in graphein.protein.config
    :type config: graphein.protein.config.ProteinMeshConfig
    :return: verts, faces, aux
    :rtype: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
    """
    from pytorch3d.io import load_obj

    if config is None:
        config = ProteinMeshConfig()

    obj_file = get_obj_file(
        pdb_code=pdb_code, pdb_file=pdb_file, out_dir=out_dir, config=config
    )
    verts, faces, aux = load_obj(obj_file)
    return verts, faces, aux


def normalize_and_center_mesh_vertices(
    verts: torch.FloatTensor,
) -> torch.FloatTensor:
    """
    We scale normalize and center the target mesh to fit in a sphere of radius 1 centered at (0,0,0).
    (scale, center) will be used to bring the predicted mesh to its original center and scale
    Note that normalizing the target mesh, speeds up the optimization but is not necessary!

    :param verts: mesh vertices
    :type verts: torch.FloatTensor
    :return: normalized and centered vertices
    :rtype: torch.FloatTensor
    """
    center = verts.mean()
    verts = verts - center
    scale = max(verts.abs().max(0)[0])
    verts = verts / scale
    return verts


def convert_verts_and_face_to_mesh(
    verts: torch.FloatTensor, faces: NamedTuple
) -> Meshes:
    """
    Converts vertices and faces into a pytorch3d.structures Meshes object.

    :param verts: vertices
    :type verts: torch.FloatTensor
    :param faces: faces
    :type faces: NamedTuple
    :return: Meshes object
    :rtype: pytorch3d.structures.Meshes
    """
    faces_idx = faces.verts_idx
    return Meshes(verts=[verts], faces=[faces_idx])


if __name__ == "__main__":
    from graphein.protein.visualisation import plot_pointcloud

    config = ProteinMeshConfig(
        pymol_commands=[
            "show surface",
            "set surface_solvent, on",
            "set solvent_radius, 10000",
        ]
    )
    verts, faces, aux = create_mesh(pdb_code="3eiy", config=config)

    trg_mesh = convert_verts_and_face_to_mesh(verts, faces)
    plot_pointcloud(trg_mesh)



"""Functions for plotting protein graphs and meshes"""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from itertools import count
from typing import List, Optional, Tuple

import matplotlib
import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import plotly.graph_objects as go
from mpl_toolkits.mplot3d import Axes3D

from graphein.utils.utils import import_message

try:
    from pytorch3d.ops import sample_points_from_meshes
except ImportError:
    import_message(
        submodule="graphein.protein.visualisation",
        package="pytorch3d",
        conda_channel="pytorch3d",
    )


def plot_pointcloud(mesh: Meshes, title: str = "") -> Axes3D:
    """
    Plots pytorch3d Meshes object as pointcloud

    :param mesh: Meshes object to plot
    :type mesh: pytorch3d.structures.meshes.Meshes
    :param title: Title of plot
    :type title: str
    :return: returns Axes3D containing plot
    :rtype: Axes3D
    """
    # Sample points uniformly from the surface of the mesh.
    points = sample_points_from_meshes(mesh, 5000)
    x, y, z = points.clone().detach().cpu().squeeze().unbind(1)
    fig = plt.figure(figsize=(5, 5))
    ax = Axes3D(fig)
    ax.scatter3D(x, z, -y)
    ax.set_xlabel("x")
    ax.set_ylabel("z")
    ax.set_zlabel("y")
    ax.set_title(title)
    ax.view_init(190, 30)
    return ax


def colour_nodes(
    G: nx.Graph, colour_map: matplotlib.colors.ListedColormap, colour_by: str
) -> List[Tuple[float, float, float, float]]:
    """
    Computes node colours based on "degree", "seq_position" or node attributes

    :param G: Graph to compute node colours for
    :type G: nx.Graph
    :param colour_map:  Colourmap to use.
    :type colour_map: matplotlib.colors.ListedColormap
    :param colour_by: Manner in which to colour nodes. If not "degree" or "seq_position", this must correspond to a node feature
    :type colour_by: str
    :return: List of node colours
    :rtype: List[Tuple[float, float, float, float]]
    """
    # get number of nodes
    n = G.number_of_nodes()

    # Get max number of edges connected to a single node
    edge_max = max([G.degree[i] for i in G.nodes()])

    # Define color range proportional to number of edges adjacent to a single node
    if colour_by == "degree":
        colors = [colour_map(G.degree[i] / edge_max) for i in G.nodes()]
    elif colour_by == "seq_position":
        colors = [colour_map(i / n) for i in range(n)]
    else:
        node_types = set(nx.get_node_attributes(G, colour_by).values())
        mapping = dict(zip(sorted(node_types), count()))
        colors = [
            colour_map(mapping[d[colour_by]] / len(node_types))
            for n, d in G.nodes(data=True)
        ]

    return colors


def colour_edges(
    G: nx.Graph,
    colour_map: matplotlib.colors.ListedColormap,
    colour_by: str = "kind",
) -> List[Tuple[float, float, float, float]]:
    """
    Computes edge colours based on the kind of bond/interaction.

    :param G: nx.Graph protein structure graph to compute edge colours from
    :type G: nx.Graph
    :param colour_map: Colourmap to use
    :type colour_map: matplotlib.colors.ListedColormap
    :param colour_by: Edge attribute to colour by. Currently only "kind" is supported
    :type colour_by: str
    :return: List of edge colours
    :rtype: List[Tuple[float, float, float, float]]
    """
    if colour_by == "kind":
        edge_types = set(
            frozenset(a) for a in nx.get_edge_attributes(G, "kind").values()
        )
        mapping = dict(zip(sorted(edge_types), count()))
        colors = [
            colour_map(
                mapping[frozenset(G.edges[i]["kind"])] / len(edge_types)
            )
            for i in G.edges()
        ]
    return colors


def plotly_protein_structure_graph(
    G: nx.Graph,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (620, 650),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 20.0,
    label_node_ids: bool = True,
    node_colour_map=plt.cm.plasma,
    edge_color_map=plt.cm.plasma,
    colour_nodes_by: str = "degree",
    colour_edges_by: str = "type",
    edge_alpha: float = 0.5,
) -> go.Figure:
    """
    Plots protein structure graph using plotly.

    :param G:  nx.Graph Protein Structure graph to plot
    :type G: nx.Graph
    :param plot_title: Title of plot, defaults to None
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to (620, 650)
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to 0.7
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes reflect degree.
    :type node_size_multiplier: float
    :param label_node_ids: bool indicating whether or not to plot node_id labels
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. "degree", "seq_position" or a node feature
    :type colour_edges_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only "kind" is supported
    :type colour_nodes_by: str
    :param edge_alpha: Controls edge transparency
    :type edge_alpha: float
    :returns: Plotly Graph Objects plot
    :rtype: go.Figure
    """

    # Get Node Attributes
    pos = nx.get_node_attributes(G, "coords")

    # Get node colours
    node_colors = colour_nodes(
        G, colour_map=node_colour_map, colour_by=colour_nodes_by
    )
    edge_colors = colour_edges(
        G, colour_map=edge_color_map, colour_by=colour_edges_by
    )

    # 3D network plot
    x_nodes = []
    y_nodes = []
    z_nodes = []
    node_sizes = []
    node_labels = []

    # Loop on the pos dictionary to extract the x,y,z coordinates of each node
    for i, (key, value) in enumerate(pos.items()):
        x_nodes.append(value[0])
        y_nodes.append(value[1])
        z_nodes.append(value[2])
        node_sizes.append(node_size_min + node_size_multiplier * G.degree[key])

        if label_node_ids:
            node_labels.append(list(G.nodes())[i])

    nodes = go.Scatter3d(
        x=x_nodes,
        y=y_nodes,
        z=z_nodes,
        mode="markers",
        marker={
            "symbol": "circle",
            "color": node_colors,
            "size": node_sizes,
            "opacity": node_alpha,
        },
        text=list(G.nodes()),
        hoverinfo="text+x+y+z",
    )

    # Loop on the list of edges to get the x,y,z, coordinates of the connected nodes
    # Those two points are the extrema of the line to be plotted
    x_edges = []
    y_edges = []
    z_edges = []

    for node_a, node_b in G.edges(data=False):
        x_edges.extend([pos[node_a][0], pos[node_b][0], None])
        y_edges.extend([pos[node_a][1], pos[node_b][1], None])
        z_edges.extend([pos[node_a][2], pos[node_b][2], None])

    axis = dict(
        showbackground=False,
        showline=False,
        zeroline=False,
        showgrid=False,
        showticklabels=False,
        title="",
    )

    edges = go.Scatter3d(
        x=x_edges,
        y=y_edges,
        z=z_edges,
        mode="lines",
        line={"color": edge_colors, "width": 10},
        text=[
            str(list(edge_type))
            for edge_type in nx.get_edge_attributes(G, "kind").values()
        ],
        hoverinfo="text",
    )

    fig = go.Figure(
        data=[nodes, edges],
        layout=go.Layout(
            title=plot_title,
            width=figsize[0],
            height=figsize[1],
            showlegend=False,
            scene=dict(
                xaxis=dict(axis),
                yaxis=dict(axis),
                zaxis=dict(axis),
            ),
            margin=dict(t=100),
        ),
    )

    return fig


def plot_protein_structure_graph(
    G: nx.Graph,
    angle: int = 30,
    plot_title: Optional[str] = None,
    figsize: Tuple[int, int] = (10, 7),
    node_alpha: float = 0.7,
    node_size_min: float = 20.0,
    node_size_multiplier: float = 20.0,
    label_node_ids: bool = True,
    node_colour_map=plt.cm.plasma,
    edge_color_map=plt.cm.plasma,
    colour_nodes_by: str = "degree",
    colour_edges_by: str = "kind",
    edge_alpha: float = 0.5,
    plot_style: str = "ggplot",
    out_path: Optional[str] = None,
    out_format: str = ".png",
) -> Axes3D:
    """
    Plots protein structure graph in Axes3D.

    :param G:  nx.Graph Protein Structure graph to plot
    :type G: nx.Graph
    :param angle:  View angle
    :type angle: int
    :param plot_title: Title of plot. Defaults to None
    :type plot_title: str, optional
    :param figsize: Size of figure, defaults to (10, 7)
    :type figsize: Tuple[int, int]
    :param node_alpha: Controls node transparency, defaults to 0.7
    :type node_alpha: float
    :param node_size_min: Specifies node minimum size, defaults to 20
    :type node_size_min: float
    :param node_size_multiplier: Scales node size by a constant. Node sizes reflect degree.
    :type node_size_multiplier: float
    :param label_node_ids: bool indicating whether or not to plot node_id labels
    :type label_node_ids: bool
    :param node_colour_map: colour map to use for nodes
    :type node_colour_map: plt.cm
    :param edge_color_map: colour map to use for edges
    :type edge_color_map: plt.cm
    :param colour_nodes_by: Specifies how to colour nodes. "degree", "seq_position" or a node feature
    :type colour_nodes_by: str
    :param colour_edges_by: Specifies how to colour edges. Currently only "kind" is supported
    :param edge_alpha: Controls edge transparency
    :param plot_style: matplotlib style sheet to use
    :param out_path: If not none, writes plot to this location
    :param out_format: Fileformat to use for plot
    :return:
    """

    # Get Node Attributes
    pos = nx.get_node_attributes(G, "coords")

    # Get node colours
    node_colors = colour_nodes(
        G, colour_map=node_colour_map, colour_by=colour_nodes_by
    )
    edge_colors = colour_edges(
        G, colour_map=edge_color_map, colour_by=colour_edges_by
    )

    # 3D network plot
    with plt.style.context(plot_style):

        fig = plt.figure(figsize=figsize)
        ax = Axes3D(fig, auto_add_to_figure=True)

        # Loop on the pos dictionary to extract the x,y,z coordinates of each node
        for i, (key, value) in enumerate(pos.items()):
            xi = value[0]
            yi = value[1]
            zi = value[2]

            # Scatter plot
            ax.scatter(
                xi,
                yi,
                zi,
                color=node_colors[i],
                s=node_size_min + node_size_multiplier * G.degree[key],
                edgecolors="k",
                alpha=node_alpha,
            )
            if label_node_ids:
                label = list(G.nodes())[i]
                ax.text(xi, yi, zi, label)

        # Loop on the list of edges to get the x,y,z, coordinates of the connected nodes
        # Those two points are the extrema of the line to be plotted
        for i, j in enumerate(G.edges()):
            x = np.array((pos[j[0]][0], pos[j[1]][0]))
            y = np.array((pos[j[0]][1], pos[j[1]][1]))
            z = np.array((pos[j[0]][2], pos[j[1]][2]))

            # Plot the connecting lines
            ax.plot(x, y, z, c=edge_colors[i], alpha=edge_alpha)

    # Set title
    ax.set_title(plot_title)
    # Set the initial view
    ax.view_init(30, angle)
    # Hide the axes
    ax.set_axis_off()
    if out_path is not None:
        plt.savefig(out_path + str(angle).zfill(3) + out_format)
        plt.close("all")

    return ax


if __name__ == "__main__":
    # TODO: Move the block here into tests.
    from graphein.protein.config import ProteinGraphConfig
    from graphein.protein.edges.atomic import (
        add_atomic_edges,
        add_bond_order,
        add_ring_status,
    )
    from graphein.protein.features.nodes.amino_acid import (
        expasy_protein_scale,
        meiler_embedding,
    )
    from graphein.protein.graphs import construct_graph

    # Test Point cloud plotting
    # v, f, a = create_mesh(pdb_code="3eiy")
    # m = convert_verts_and_face_to_mesh(v, f)
    # plot_pointcloud(m, "Test")
    # TEST PROTEIN STRUCTURE GRAPH PLOTTING
    configs = {
        "granularity": "atom",
        "keep_hets": False,
        "deprotonate": True,
        "insertions": False,
        "verbose": False,
    }

    config = ProteinGraphConfig(**configs)
    config.edge_construction_functions = [
        add_atomic_edges,
        add_ring_status,
        add_bond_order,
    ]

    config.node_metadata_functions = [meiler_embedding, expasy_protein_scale]
    # g = construct_graph(config=config, pdb_path="../../examples/pdbs/1a1e.pdb", pdb_code="1a1e")

    g = construct_graph(
        config=config, pdb_path="../examples/pdbs/3eiy.pdb", pdb_code="3eiy"
    )
    print(nx.info(g))

    p = plotly_protein_structure_graph(
        g,
        30,
        (1000, 2000),
        colour_nodes_by="element_symbol",
        colour_edges_by="kind",
        label_node_ids=False,
    )

    p.show()



"""Featurization functions for graph nodes using DSSP predicted features."""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Charlie Harris
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
from typing import Any, Dict, Optional

import networkx as nx
import pandas as pd
from Bio.Data.IUPACData import protein_letters_1to3
from Bio.PDB.DSSP import dssp_dict_from_pdb_file, residue_max_acc

from graphein.protein.utils import download_pdb

DSSP_COLS = [
    "chain",
    "resnum",
    "icode",
    "aa",
    "ss",
    "asa",
    "phi",
    "psi",
    "dssp_index",
    "NH_O_1_relidx",
    "NH_O_1_energy",
    "O_NH_1_relidx",
    "O_NH_1_energy",
    "NH_O_2_relidx",
    "NH_O_2_energy",
    "O_NH_2_relidx",
    "O_NH_2_energy",
]

DSSP_SS = ["H", "B", "E", "G", "I", "T", "S"]


def parse_dssp_df(dssp: Dict[str, Any]) -> pd.DataFrame:
    """
    Parse DSSP output to DataFrame

    :param dssp: Dictionary containing DSSP output
    :type dssp: Dict[str, Any]
    :return: pd.Dataframe containing parsed DSSP output
    :rtype: pd.DataFrame
    """
    appender = []
    for k in dssp[1]:
        to_append = []
        y = dssp[0][k]
        chain = k[0]
        residue = k[1]
        # het = residue[0]
        resnum = residue[1]
        icode = residue[2]
        to_append.extend([chain, resnum, icode])
        to_append.extend(y)
        appender.append(to_append)

    return pd.DataFrame.from_records(appender, columns=DSSP_COLS)


def process_dssp_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Processes a DSSP DataFrame to make indexes align with node IDs

    :param df: pd.DataFrame containing the parsed output from DSSP.
    :type df: pd.DataFrame
    :return: pd.DataFrame with node IDs
    :rtype: pd.DataFrame
    """

    # Convert 1 letter aa code to 3 letter
    amino_acids = df["aa"].tolist()

    for i, amino_acid in enumerate(amino_acids):
        amino_acids[i] = protein_letters_1to3[amino_acid].upper()
    df["aa"] = amino_acids

    # Construct node IDs
    node_ids = []

    for i, row in df.iterrows():
        node_id = row["chain"] + ":" + row["aa"] + ":" + str(row["resnum"])
        node_ids.append(node_id)
    df["node_id"] = node_ids

    df.set_index("node_id", inplace=True)

    return df


def add_dssp_df(G: nx.Graph, dssp_config: Optional[DSSPConfig]) -> nx.Graph:
    """
    Construct DSSP dataframe and add as graph level variable to protein graph

    :param G: Input protein graph
    :param G: nx.Graph
    :param dssp_config: DSSPConfig object. Specifies which executable to run. Located in graphein.protein.config
    :type dssp_config: DSSPConfig, optional
    :return: Protein graph with DSSP dataframe added
    :rtype: nx.Graph
    """

    config = G.graph["config"]
    pdb_id = G.graph["pdb_id"]

    # TODO - Check for DSSP installation

    # Check for existence of pdb file. If not, download it.
    if not os.path.isfile(config.pdb_dir / pdb_id):
        pdb_file = download_pdb(config, pdb_id)
    else:
        pdb_file = config.pdb_dir + pdb_id + ".pdb"

    # Extract DSSP executable
    executable = dssp_config.executable

    if config.verbose:
        print(f"Using DSSP executable '{executable}'")

    # Run DSSP
    dssp_dict = dssp_dict_from_pdb_file(pdb_file, DSSP=executable)
    dssp_dict = parse_dssp_df(dssp_dict)
    dssp_dict = process_dssp_df(dssp_dict)

    if config.verbose:
        print(dssp_dict)

    # Assign DSSP Dict
    G.graph["dssp_df"] = dssp_dict

    return G


def add_dssp_feature(G: nx.Graph, feature: str) -> nx.Graph:
    """
    Adds add_dssp_feature specified amino acid feature as calculated
    by DSSP to every node in a protein graph
    :param G: Protein structure graph to add dssp feature to
    :param feature: string specifying name of DSSP feature to add:
    "chain",
    "resnum",
    "icode",
    "aa",
    "ss",
    "asa",
    "phi",
    "psi",
    "dssp_index",
    "NH_O_1_relidx",
    "NH_O_1_energy",
    "O_NH_1_relidx",
    "O_NH_1_energy",
    "NH_O_2_relidx",
    "NH_O_2_energy",
    "O_NH_2_relidx",
    "O_NH_2_energy",

    These names parse_dssp_df accessible in the DSSP_COLS list
    :param G: Protein Graph to add features to
    :type G: nx.Graph
    :return: Protein structure graph with DSSP feature added to nodes
    :rtype: nx.Graph
    """
    if "dssp_df" not in G.graph:
        G = add_dssp_df(G, G.graph["config"].dssp_config)

    config = G.graph["config"]
    dssp_df = G.graph["dssp_df"]

    # Change to not allow for atom granuarlity?
    if config.granularity == "atom":
        # TODO confirm below is not needed and remove
        """
        # If granularity is atom, apply residue feature to every atom
        for n in G.nodes():
            residue = n.split(":")
            residue = residue[0] + ":" + residue[1] + ":" + residue[2]

            G.nodes[n][feature] = dssp_df.loc[residue, feature]
        """
        raise NameError(
            f"DSSP residue features ({feature}) \
            cannot be added to atom granularity graph"
        )

    else:
        nx.set_node_attributes(G, dict(dssp_df[feature]), feature)

    if config.verbose:
        print("Added " + feature + " features to graph nodes")

    return G


def rsa(G: nx.Graph) -> nx.Graph:
    """
    Adds RSA (relative solvent accessibility) of each residue in protein graph
    as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with rsa values added
    :rtype: nx.Graph
    """

    # Calculate RSA
    dssp_df = G.graph["dssp_df"]
    dssp_df["max_acc"] = dssp_df["aa"].map(residue_max_acc["Sander"].get)
    dssp_df[["asa", "max_acc"]] = dssp_df[["asa", "max_acc"]].astype(float)
    dssp_df["rsa"] = dssp_df["asa"] / dssp_df["max_acc"]

    G.graph["dssp_df"] = dssp_df

    return add_dssp_feature(G, "rsa")


def asa(G: nx.Graph) -> nx.Graph:
    """
    Adds ASA of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with asa values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "asa")


def phi(G: nx.Graph) -> nx.Graph:
    """
    Adds phi-angles of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with phi-angles values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "phi")


def psi(G: nx.Graph) -> nx.Graph:
    """
    Adds psi-angles of each residue in protein graph as calculated by DSSP.

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with psi-angles values added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "psi")


def secondary_structure(G: nx.Graph) -> nx.Graph:
    """
    Adds secondary structure of each residue in protein graph
    as calculated by DSSP in the form of a string

    :param G: Input protein graph
    :type G: nx.Graph
    :return: Protein graph with secondary structure added
    :rtype: nx.Graph
    """
    return add_dssp_feature(G, "ss")



"""Base Config object for use with Protein Graph Construction"""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

from pathlib import Path
from typing import Callable, List, Literal, Optional, Union

from pydantic import BaseModel

from graphein.protein.edges.distance import add_peptide_bonds
from graphein.protein.features.nodes.amino_acid import meiler_embedding


class DSSPConfig(BaseModel):
    executable: str = "mkdssp"


class GetContactsConfig(BaseModel):
    """
    Config object for parameters relating to running GetContacts. GetContacts is an optional dependency from which intramolecular interactions can be computed and used as edges in the graph.

    :param get_contacts_path: Path to GetContacts installation
    :type get_contacts_path: pathlib.Path
    :param contacts_dir: Path to store output of GetContacts
    :type contacts_dir: pathlib.Path
    :param pdb_dir: Path to PDB files to be used to compute intramolecular interactions.
    :type pdb_dir: pathlib.Path
    :param granularity: Specifies the node types of the graph, defaults to "CA" for alpha-carbons as nodes. Other options are "CB" (beta-carbon), "atom" for all-atom graphs, and "centroid" for nodes positioned as residue centroids.
    :type granularity: str
    """

    get_contacts_path: Path = Path(
        "/Users/arianjamasb/github/getcontacts/"
    ).resolve()
    contacts_dir: Path = Path("../examples/contacts/").resolve()
    pdb_dir: Path = Path("../examples/pdbs/").resolve()
    granularity: str = "CA"


GraphAtoms = Literal[
    "N",
    "CA",
    "C",
    "O",
    "CB",
    "OG",
    "CG",
    "CD1",
    "CD2",
    "CE1",
    "CE2",
    "CZ",
    "OD1",
    "ND2",
    "CG1",
    "CG2",
    "CD",
    "CE",
    "NZ",
    "OD2",
    "OE1",
    "NE2",
    "OE2",
    "OH",
    "NE",
    "NH1",
    "NH2",
    "OG1",
    "SD",
    "ND1",
    "SG",
    "NE1",
    "CE3",
    "CZ2",
    "CZ3",
    "CH2",
    "OXT",
]
GranularityOpts = Literal["atom", "centroids"]


class ProteinGraphConfig(BaseModel):
    """
    Config Object for Protein Structure Graph Construction

    :param granularity: Controls the granularity of the graph construction. "atom" builds an atomic-scale graph where
        nodes are constituent atoms. Residue-level graphs can be build by specifying which constituent atom should
        represent node positions (see graphein.protein.config.GraphAtoms). Additionally, "centroids" can be specified to
        compute the centre of gravity for a given atom. Defaults to "CA" (alpha-Carbon)
    :type granularity: str (Union[graphein.protein.config.GraphAtoms, graphein.protein.config.GranularityOpts])
    :param keep_hets: Controls whether or not heteroatoms are removed from the PDB file. These are typically modified
        residules, bound ligands, crystallographic adjuvants, ions or water molecules.
    :type keep_hets: bool
    :param insertions: Controls whether or not insertions are allowed.
    :type insertions: bool
    :param pdb_dir: Specifies path to download protein structures into.
    :type pdb_dir: pathlib.Path
    :param verbose: Specifies verbosity of graph creation process.
    :type verbose: bool
    :param exclude_waters: Specifies whether or not water molecules are excluded from the structure
    :type excluded_waters: bool
    :param deprotonate: Specifies whether or not to remove H atoms from the graph.
    :type deprotonate: bool
    :param protein_df_processing_functions: List of functions that take a pd.DataFrame and return a pd.DataFrame.
        This allows users to define their own series of processing functions for the protein structure DataFrame and
        override the default sequencing of processing steps provided by Graphein. We refer users to our low-level API
        tutorial for more details.
    :type protein_df_processing_functions: Optional[List[Callable]]

    :param edge_construction_functions: List of functions that take an nx.Graph and return an nx.Graph with desired
        edges added. Prepared edge constructions can be found in graphein.protein.edges
    :type edge_construction_functions: List[Callable]

    :param node_metadata_functions: List of functions that taken an nx.Graph
    :type node_metadata_functions: List[Callable], optional

    :param edge_metadata_functions: List of functions that take an
    :type edge_metadata_functions: List[Callable], optional

    :param graph_metadata_functions: List of functions that take an nx.Graph and return an nx.Graph with added
        graph-level features and metadata.
    :type graph_metadata_functions: List[Callable], optional

    :param get_contacts_config: Config object containing parameters for running GetContacts for computing intramolecular
        contact-based edges. Defaults to None.
    :type get_contacts_config: GetContactsConfig, optional
    :param dssp_config: Config Object containing reference to executable. Defaults to None.
    :type dssp_config: DSSPConfig, optional
    """

    granularity: Union[GraphAtoms, GranularityOpts] = "CA"
    keep_hets: bool = False
    insertions: bool = False
    pdb_dir: Path = Path(
        "../examples/pdbs/"
    )  # Also suggest to avoid hard-coding paths if possible!
    verbose: bool = False
    exclude_waters: bool = True
    deprotonate: bool = False

    # Graph construction functions
    protein_df_processing_functions: Optional[List[Callable]] = None
    edge_construction_functions: List[Union[Callable, str]] = [
        add_peptide_bonds
    ]
    node_metadata_functions: Optional[List[Union[Callable, str]]] = [
        meiler_embedding
    ]
    edge_metadata_functions: Optional[List[Union[Callable, str]]] = None
    graph_metadata_functions: Optional[List[Callable]] = None

    # External Dependency configs
    get_contacts_config: Optional[GetContactsConfig] = None
    dssp_config: Optional[DSSPConfig] = None


class ProteinMeshConfig(BaseModel):
    """
    Config object for parameters relating to ProteinMeshConfig Mesh construction with PyMol

    :param pymol_command_line_options: List of CLI args for running PyMol.
        See: https://pymolwiki.org/index.php/Command_Line_Options. Defaults to "-cKq"
    :type pymol_command_line_options: str, optional
    :param pymol_commands: List of Commands passed to PyMol in surface construction.
    :type pymol_commands: List[str], optional
    """

    pymol_command_line_options: Optional[str] = "-cKq"
    pymol_commands: Optional[List[str]] = ["show surface"]



from __future__ import annotations

from typing import List, Literal, Optional, TypeVar

import networkx as nx
import numpy as np
import torch

from graphein.utils.utils import import_message

try:
    from torch_geometric.data import Data
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="torch_geometric",
        pip_install=True,
        conda_channel="rusty1s",
    )

try:
    import dgl
except ImportError:
    import_message(
        submodule="graphein.ml.conversion",
        package="dgl",
        pip_install=True,
        conda_channel="dglteam",
    )


SUPPORTED_FORMATS = ["nx", "pyg", "dgl"]
SUPPORTED_VERBOSITY = ["gnn", "default", "all_info"]


class GraphFormatConvertor:
    """
    Provides conversion utilities between NetworkX Graphs and geometric deep learning library destination formats.
    Currently, we provide support for converstion from nx.Graphs to dgl.DGLGraph and pytorch_geometric.Data. Supported conversion
    formats can be retrieved from graphein.ml.conversion.SUPPORTED_FORMATS.

    :param src_format: The type of graph you'd like to convert from. Supported formats are available in graphein.ml.conversion.SUPPORTED_FORMATS
    :type src_format: Literal["nx", "pyg", "dgl"]
    :param dst_format: The type of graph format you'd like to convert to. Supported formats are available in:
        graphein.ml.conversion.SUPPORTED_FORMATS
    :type dst_format:  Literal["nx", "pyg", "dgl"]
    :param verbose: Select from "gnn", "default", "all_info" to determine how much information is preserved (features)
        as some are unsupported by various downstream frameworks
    :type verbose: graphein.ml.conversion.SUPPORTED_VERBOSITY
    :param columns: List of columns in the node features to retain
    :type columns: List[str], optional
    """

    def __init__(
        self,
        src_format: str,
        dst_format: str,
        verbose: SUPPORTED_VERBOSITY = "gnn",
        columns: Optional[List[str]] = None,
    ):
        if (src_format not in SUPPORTED_FORMATS) or (
            dst_format not in SUPPORTED_FORMATS
        ):
            raise ValueError(
                "Please specify from supported format, "
                + "/".join(SUPPORTED_FORMATS)
            )
        self.src_format = src_format
        self.dst_format = dst_format

        # supported_verbose_format = ["gnn", "default", "all_info"]
        if (columns is None) and (verbose not in SUPPORTED_VERBOSITY):
            raise ValueError(
                "Please specify the supported verbose mode ("
                + "/".join(SUPPORTED_VERBOSITY)
                + ") or specify column names!"
            )

        if columns is None:
            if verbose == "gnn":
                columns = [
                    "edge_index",
                    "coords",
                    "dist_mat",
                    "name",
                    "node_id",
                ]
            elif verbose == "default":
                columns = [
                    "b_factor",
                    "chain_id",
                    "coords",
                    "dist_mat",
                    "edge_index",
                    "kind",
                    "name",
                    "node_id",
                    "residue_name",
                ]
            elif verbose == "all_info":
                columns = [
                    "atom_type",
                    "b_factor",
                    "chain_id",
                    "chain_ids",
                    "config",
                    "coords",
                    "dist_mat",
                    "edge_index",
                    "element_symbol",
                    "kind",
                    "name",
                    "node_id",
                    "node_type",
                    "pdb_df",
                    "raw_pdb_df",
                    "residue_name",
                    "residue_number",
                    "rgroup_df",
                    "sequence_A",
                    "sequence_B",
                ]
        self.columns = columns

        self.type2form = {
            "atom_type": "str",
            "b_factor": "float",
            "chain_id": "str",
            "coords": "np.array",
            "dist_mat": "np.array",
            "element_symbol": "str",
            "node_id": "str",
            "residue_name": "str",
            "residue_number": "int",
            "edge_index": "torch.tensor",
            "kind": "str",
        }

    def convert_nx_to_dgl(self, G: nx.Graph) -> dgl.DGLGraph:
        """
        Converts NetworkX graph to DGL

        :param G: nx.Graph to convert to DGLGraph
        :type G: nx.Graph
        :return: DGLGraph object version of input NetworkX graph
        :rtype: dgl.DGLGraph
        """
        g = dgl.DGLGraph()
        node_id = [n for n in G.nodes()]
        G = nx.convert_node_labels_to_integers(G)

        ## add node level feat

        node_dict = {}
        for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    node_dict[str(key)] = (
                        [value] if i == 0 else node_dict[str(key)] + [value]
                    )

        string_dict = {}
        node_dict_transformed = {}
        for i, j in node_dict.items():
            if i == "coords":
                node_dict_transformed[i] = torch.Tensor(np.asarray(j)).type(
                    "torch.FloatTensor"
                )
            elif i == "dist_mat":
                node_dict_transformed[i] = torch.Tensor(
                    np.asarray(j[0].values)
                ).type("torch.FloatTensor")
            elif self.type2form[i] == "str":
                string_dict[i] = j
            elif self.type2form[i] == "float":
                node_dict_transformed[i] = torch.Tensor(np.array(j))
            elif self.type2form[i] == "int":
                node_dict_transformed[i] = torch.Tensor(np.array(j))

        g.add_nodes(
            len(node_id),
            node_dict_transformed,
        )

        edge_dict = {}
        edge_index = torch.LongTensor(list(G.edges)).t().contiguous()

        # add edge level features
        for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    edge_dict[str(key)] = (
                        list(value)
                        if i == 0
                        else edge_dict[str(key)] + list(value)
                    )

        edge_transform_dict = {}
        for i, j in node_dict.items():
            if self.type2form[i] == "str":
                string_dict[i] = j
            elif self.type2form[i] == "float":
                edge_transform_dict[i] = torch.Tensor(np.array(j))
            elif self.type2form[i] == "int":
                edge_transform_dict[i] = torch.Tensor(np.array(j))

        g.add_edges(edge_index[0], edge_index[1], edge_transform_dict)

        # add graph level features
        graph_dict = {}
        for i, feat_name in enumerate(G.graph):
            if str(feat_name) in self.columns:
                graph_dict[str(feat_name)] = [G.graph[feat_name]]

        return g

    def convert_nx_to_pyg(self, G: nx.Graph) -> Data:
        """
        Converts NetworkX graph to pytorch_geometric.data.Data object. Requires Torch Geometric to be installed.

        :param G: nx.Graph to convert to PyTorch Geometric
        :type G: nx.Graph
        :return: Data object containing networkx graph data
        :rtype: pytorch_geometric.data.Data
        """

        # Initialise dict used to construct Data object
        data = {}

        # Assign node ids as a feature
        data["node_id"] = [n for n in G.nodes()]
        G = nx.convert_node_labels_to_integers(G)

        # Construct Edge Index
        edge_index = torch.LongTensor(list(G.edges)).t().contiguous()

        # Add node features
        for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    data[str(key)] = (
                        [value] if i == 0 else data[str(key)] + [value]
                    )

        # Add edge features
        for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
            for key, value in feat_dict.items():
                if str(key) in self.columns:
                    data[str(key)] = (
                        list(value) if i == 0 else data[str(key)] + list(value)
                    )

        # Add graph-level features
        for i, feat_name in enumerate(G.graph):
            if str(feat_name) in self.columns:
                data[str(feat_name)] = [G.graph[feat_name]]

        if "edge_index" in self.columns:
            data["edge_index"] = edge_index.view(2, -1)

        data = Data.from_dict(data)
        data.num_nodes = G.number_of_nodes()
        return data

    @staticmethod
    def convert_nx_to_nx(G: nx.Graph) -> nx.Graph:
        """
        Converts NetworkX Graph to NetworkX graph object. Redundant - returns itself

        :param G: NetworkX Graph
        :type G: nx.Graph
        :return: NetworkX Graph
        :rtype: nx.Graph
        """
        return G

    @staticmethod
    def convert_dgl_to_nx(G: dgl.DGLGraph) -> nx.Graph:
        """
        Converts a DGL Graph (dgl.DGLGraph) to a NetworkX (nx.Graph) object. Preservers node and edge attributes.

        :param G: dgl.DGLGraph to convert to NetworkX
        :type G: dgl.DGLGraph
        :return: NetworkX graph object
        :rtype: nx.Graph
        """
        node_attrs = G.node_attr_schemes().keys()
        edge_attrs = G.edge_attr_schemes().keys()
        nx_g = dgl.to_networkx(G, node_attrs, edge_attrs)
        return nx_g

    @staticmethod
    def convert_pyg_to_nx(G: Data) -> nx.Graph:
        """Converts PyTorch Geometric Data object to NetworkX graph

        :param G: Pytorch Geometric Data
        :type G: torch_geometric.data.Data
        :returns: NetworkX graph version
        :rtype: nx.Graph
        """
        return torch_geometric.utils.to_networkx(G)

    def __call__(self, G: nx.Graph):
        nx_g = eval("self.convert_" + self.src_format + "_to_nx(G)")
        dst_g = eval("self.convert_nx_to_" + self.dst_format + "(nx_g)")
        return dst_g


def convert_nx_to_pyg_data(G: nx.Graph) -> Data:
    # Initialise dict used to construct Data object
    data = {}

    # Assign node ids as a feature
    data["node_id"] = [n for n in G.nodes()]
    G = nx.convert_node_labels_to_integers(G)

    # Construct Edge Index
    edge_index = torch.LongTensor(list(G.edges)).t().contiguous()

    # Add node features
    for i, (_, feat_dict) in enumerate(G.nodes(data=True)):
        for key, value in feat_dict.items():
            data[str(key)] = [value] if i == 0 else data[str(key)] + [value]

    # Add edge features
    for i, (_, _, feat_dict) in enumerate(G.edges(data=True)):
        for key, value in feat_dict.items():
            data[str(key)] = (
                list(value) if i == 0 else data[str(key)] + list(value)
            )

    # Add graph-level features
    for i, feat_name in enumerate(G.graph):
        data[str(feat_name)] = [G.graph[feat_name]]

    data["edge_index"] = edge_index.view(2, -1)
    data = Data.from_dict(data)
    data.num_nodes = G.number_of_nodes()

    return data


if __name__ == "__main__":
    from graphein.protein.config import ProteinGraphConfig
    from graphein.protein.graphs import construct_graph

    g = construct_graph(pdb_code="3eiy", config=ProteinGraphConfig())
    assert type(g) is nx.Graph

    # print(SUPPORTED_FORMATS)

    convertor = GraphFormatConvertor(
        src_format="nx", dst_format="pyg", verbose="gnn"
    )
    pyg = convertor(g)
    assert type(pyg) is torch_geometric.data.Data



"""Protein graph construction module."""
from .config import *
from .edges import *
from .features import *
from .graphs import *
from .resi_atoms import *
from .utils import *
from .visualisation import plot_protein_structure_graph

try:
    from .meshes import *
    from .visualisation import plot_pointcloud
except ImportError:
    pass



"""Utilities for working with graph objects"""
# Graphein
# Author: Arian Jamasb <arian@jamasb.io>, Eric Ma
# License: MIT
# Project Website: https://github.com/a-r-j/graphein
# Code Repository: https://github.com/a-r-j/graphein
from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Any, Callable, Iterable, List, Optional

import networkx as nx
import numpy as np
import pandas as pd
import xarray as xr
from Bio.Data.IUPACData import protein_letters_3to1
from yaml import unsafe_load


def onek_encoding_unk(
    x: Iterable[Any], allowable_set: List[Any]
) -> List[bool]:
    """
    Function for perfroming one hot encoding

    :param x: value to one-hot
    :type x: Iterable[Any]
    :param allowable_set: set of options to encode
    :type allowable_set: List[Any]
    :return: one-hot encoding as list
    :rtype: List[bool]
    """
    # if x not in allowable_set:
    #    x = allowable_set[-1]
    return [x == s for s in allowable_set]


def filter_dataframe(df: pd.DataFrame, funcs: List[Callable]) -> pd.DataFrame:
    """ """
    for func in funcs:
        df = func(df)
    return df


def annotate_graph_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates graph with graph-level metadata

    :param G: Graph on which to add graph-level metadata to
    :type G: nx.Graph
    :param funcs: List of graph metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph on which with node metadata added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def annotate_edge_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates Graph edges with edge metadata

    :param G: Graph to add edge metadata to
    :type G: nx.Graph
    :param funcs: List of edge metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph with edge metadata added
    :rtype: nx.Graph
    """
    for func in funcs:
        for u, v, d in G.edges(data=True):
            func(u, v, d)
    return G


def annotate_node_metadata(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates nodes with metadata

    :param G: Graph to add node metadata to
    :type G: nx.Graph
    :param funcs: List of node metadata annotation functions
    :type funcs: List[Callable]
    :return: Graph with node metadata added
    :rtype: nx.Graph
    """

    for func in funcs:
        for n, d in G.nodes(data=True):
            func(n, d)
    return G


def annotate_node_features(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Annotates nodes with features data. Note: passes whole graph to function.

    :param G: Graph to add node features to
    :type G: nx.Graph
    :param funcs: List of node feature annotation functions
    :type funcs: List[Callable]
    :return: Graph with node features added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def compute_edges(G: nx.Graph, funcs: List[Callable]) -> nx.Graph:
    """
    Computes edges for an Graph from a list of edge construction functions

    :param G: Graph to add features to
    :type G: nx.Graph
    :param funcs: List of edge construction functions
    :type funcs: List[Callable]
    :return: Graph with edges added
    :rtype: nx.Graph
    """
    for func in funcs:
        func(G)
    return G


def generate_feature_dataframe(
    G: nx.Graph,
    funcs: List[Callable],
    return_array=False,
) -> pd.DataFrame:
    """
    Return a pandas DataFrame representation of node metadata.

    `funcs` has to be list of callables whose signature is

        f(n, d) -> pd.Series

    where `n` is the graph node,
    `d` is the node metadata dictionary.
    The function must return a pandas Series whose name is the node.

    Example function:

    ```python
    def x_vec(n: Hashable, d: Dict[Hashable, Any]) -> pd.Series:
        return pd.Series({"x_coord": d["x_coord"]}, name=n)
    ```

    One fairly strong assumption is that each func
    has all the information it needs to act
    stored on the metadata dictionary.
    If you need to reference an external piece of information,
    such as a dictionary to look up values,
    set up the function to accept the dictionary,
    and use `functools.partial`
    to "reduce" the function signature to just `(n, d)`.
    An example below:

    ```python
    from functools import partial
    def get_molweight(n, d, mw_dict):
        return pd.Series({"mw": mw_dict[d["amino_acid"]]}, name=n)

    mw_dict = {"PHE": 165, "GLY": 75, ...}
    get_molweight_func = partial(get_molweight, mw_dict=mw_dict)

    generate_feature_dataframe(G, [get_molweight_func])
    ```

    The `name=n` piece is important;
    the `name` becomes the row index in the resulting dataframe.

    The series that is returned from each function
    need not only contain one key-value pair.
    You can have two or more, and that's completely fine;
    each key becomes a column in the resulting dataframe.

    A key design choice: We default to returning DataFrames,
    to make inspecting the data easy,
    but for consumption in tensor libraries,
    you can turn on returning a NumPy array
    by switching `return_array` to True.

    ## Parameters

    - `G`: A NetworkX-compatible graph object.
    - `funcs`: A list of functions.
    - `return_array`: Whether or not to return
        a NumPy array version of the data.
        Useful for consumption in tensor libs, like PyTorch or JAX.

    ## Returns

    - A pandas DataFrame.
    """
    matrix = []
    for n, d in G.nodes(data=True):
        series = []
        for func in funcs:
            res = func(n, d)
            if res.name != n:
                raise NameError(
                    f"function {func.__name__} returns a series "
                    "that is not named after the node."
                )
            series.append(res)
        matrix.append(pd.concat(series))

    df = pd.DataFrame(matrix)
    if return_array:
        return df.values
    return df


def format_adjacency(G: nx.Graph, adj: np.ndarray, name: str) -> xr.DataArray:
    """
    Format adjacency matrix nicely.

    Intended to be used when computing an adjacency-like matrix
    off a graph object G.
    For example, in defining a func:

    ```python
    def my_adj_matrix_func(G):
        adj = some_adj_func(G)
        return format_adjacency(G, adj, "xarray_coord_name")
    ```

    ## Assumptions

    1. `adj` should be a 2D matrix of shape (n_nodes, n_nodes)
    1. `name` is something that is unique amongst all names used
    in the final adjacency tensor.

    ## Parameters

    - `G`: NetworkX-compatible Graph
    - `adj`: 2D numpy array
    - `name`: A unique name for the kind of adjacency matrix
        being constructed.
        Gets used in xarray as a coordinate in the "name" dimension.

    ## Returns

    - An XArray DataArray of shape (n_nodes, n_nodes, 1)
    """
    expected_shape = (len(G), len(G))
    if adj.shape != expected_shape:
        raise ValueError(
            "Adjacency matrix is not shaped correctly, "
            f"should be of shape {expected_shape}, "
            f"instead got shape {adj.shape}."
        )
    adj = np.expand_dims(adj, axis=-1)
    nodes = list(G.nodes())
    return xr.DataArray(
        adj,
        dims=["n1", "n2", "name"],
        coords={"n1": nodes, "n2": nodes, "name": [name]},
    )


def generate_adjacency_tensor(
    G: nx.Graph, funcs: List[Callable], return_array=False
) -> xr.DataArray:
    """
    Generate adjacency tensor for a graph.

    Uses the collection of functions in `funcs`
    to build an xarray DataArray
    that houses the resulting "adjacency tensor".

    A key design choice:
    We default to returning xarray DataArrays,
    to make inspecting the data easy,
    but for consumption in tensor libraries,
    you can turn on returning a NumPy array
    by switching `return_array` to True.

    ## Parameters

    - G: NetworkX Graph.
    - funcs: A list of functions that take in G
        and return an xr.DataArray

    ## Returns
    - xr.DataArray,
        which is of shape (n_nodes, n_nodes, n_funcs).
    """
    mats = []
    for func in funcs:
        mats.append(func(G))
    da = xr.concat(mats, dim="name")
    if return_array:
        return da.data
    return da


def protein_letters_3to1_all_caps(amino_acid: str) -> str:
    """
    Converts capitalised 3 letter amino acid code to single letter. Not provided in default biopython.

    :param amino_acid: Capitalised 3-letter amino acid code (eg. "GLY")
    :type amino_acid: str
    :returns: Single-letter amino acid code
    :rtype: str
    """
    amino_acid = amino_acid[0] + amino_acid[1:].lower()
    one_letter_code = protein_letters_3to1[amino_acid]
    return one_letter_code


def import_message(
    submodule: str,
    package: str,
    conda_channel: Optional[str] = None,
    pip_install: bool = False,
):
    """
    Return warning if package is not found.
    Generic message for indicating to the user when a function relies on an
    optional module / package that is not currently installed. Includes
    installation instructions. Typically used in conjunction without optional featurisation libraries

    :param submodule: graphein submodule that needs an external dependency.
    :type submodule: str
    :param package: External package this submodule relies on.
    :type package: str
    :param conda_channel: Conda channel package can be installed from, if at all. Defaults to None
    :type conda_channel: str, optional
    :param pip_install: Whether package can be installed via pip. Defaults to False
    :type pip_install: bool
    """
    is_conda = os.path.exists(os.path.join(sys.prefix, "conda-meta"))
    installable = True
    if is_conda:
        if conda_channel is None:
            installable = False
            installation = f"{package} cannot be installed via conda"
        else:
            installation = f"conda install -c {conda_channel} {package}"
    else:
        if pip_install:
            installation = f"pip install {package}"
        else:
            installable = False
            installation = f"{package} cannot be installed via pip"

    print(
        f"To use the Graphein submodule {submodule}, you need to install "
        f"{package}."
    )
    print()
    if installable:
        print("To do so, use the following command:")
        print()
        print(f"    {installation}")
    else:
        print(f"{installation}")


def parse_config(path: Path):
    """
    Parses a yaml configuration file into a config object
    :param path: Path to configuration file
    """
    with open(path) as file:
        config_dict = unsafe_load(file)

    print(config_dict)
    if config_dict["mode"] == "protein_graph":
        return parse_protein_graph_config(config_dict)
    elif config_dict["mode"] == "protein_mesh":
        raise NotImplementedError
    elif config_dict["mode"] == "rna":
        raise NotImplementedError
    elif config_dict["mode"] == "ppi":
        raise NotImplementedError


def parse_protein_graph_config(config_dict):
    from graphein.protein.config import ProteinGraphConfig

    config = ProteinGraphConfig(**config_dict)
    print(config)
    return config


def parse_dssp_config(config_dict):
    raise NotImplementedError



from typing import Any, Callable, List

import networkx as nx
import numpy as np


def compute_feature_over_chains(
    G: nx.Graph, func: Callable, feature_name: str
) -> nx.Graph:
    """
    Computes a sequence featurisation function over the chains in a graph

    :param G: nx.Graph protein structure graph to featurise
    :type G: nx.Graph
    :param func: Sequence featurisation function
    :type func: Callable
    :param feature_name: name of added feature
    :type feature_name: str
    :return: Graph with added features of the form G.graph[f"{feature_name}_{chain_id}"]
    :rtype: nx.Graph
    """
    for c in G.graph["chain_ids"]:
        G.graph[f"{feature_name}_{c}"] = func(G.graph[f"sequence_{c}"])
        """
        feat = func(G.graph[f"sequence_{c}"])

        if out_type == "series":
            feat = pd.Series(feat)
        elif out_type == "np":
            raise NotImplementedError

        G.graph[f"{feature_name}_{c}"] = feat
        """
    return G


def aggregate_feature_over_chains(
    G: nx.Graph, feature_name: str, aggregation_type: str
) -> nx.Graph:
    """
    Performs aggregation of a given feature over chains in a graph to produce an aggregated value

    :param G: nx.Graph protein structure graph
    :type G: nx.Graph
    :param feature_name: Name of feature to aggregate
    :type feature_name: str
    :param aggregation_type: Type of aggregation to perform (min/max/mean/sum)
    :type aggregation_type: str
    :return: Graph with new aggregated feature
    :rtype: nx.Graph
    """

    if aggregation_type == "max":
        func = np.max
    elif aggregation_type == "min":
        func = np.min
    elif aggregation_type == "mean":
        func = np.mean
    elif aggregation_type == "sum":
        func = np.sum
    else:
        raise ValueError(
            f"Unsupported aggregator: {aggregation_type}. Please use min, max, mean, sum"
        )

    G.graph[f"{feature_name}_{aggregation_type}"] = func(
        [G.graph[f"{feature_name}_{c}"] for c in G.graph["chain_ids"]]
    )
    return G


def sequence_to_ngram(sequence: str, N: int) -> List[str]:
    """
    Chops a sequence into overlapping N-grams (substrings of length N)

    :param sequence: str Sequence to convert to N-garm
    :type sequence: str
    :param N: Length ofN-grams (int)
    :type N: int
    :return: List of n-grams
    :rtype: List[str]
    """
    return [sequence[i : i + N] for i in range(len(sequence) - N + 1)]


def subset_by_node_feature_value(
    G: nx.Graph, feature_name: str, feature_value: Any
) -> nx.Graph:
    """
    Extracts a subgraph from a protein structure graph based on nodes with a certain feature value

    :param G: nx.Graph protein structure graph to extract a subgraph from
    :type G: nx.Graph
    :param feature_name: Name of feature to base subgraph extraction from
    :type feature_name: str
    :param feature_value: Value of feature to select
    :type feature_value: Any
    :return: Subgraph of G based on nodes with a given feature value
    :rtype: nx.Graph
    """
    node_list = []
    for n, d in G.nodes(data=True):

        if d[feature_name] == feature_value:
            node_list.append(n)

    return G.subgraph(node_list)
